%{
/* small.md -- Machine description for elks small model using NASM
 *
 * Written by Eric Korpela, 1997, 1998
 *
 * This machine description file is free software; you can redistrubute it 
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version
 *
 * This file is distributed in the home that it will be useful, but WITHOUT
 * ANY WARRANTY; without event the implied warranty of MERCHANABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  see the GNU General Public License
 * for more details.
 *
 * First release 1998 May 13.
 */

enum { CX=0, BX=1, DX=2, AX=3, DI=4, SI=5, ES=6, DS=7 };
#include "c.h"
static char *highbyte[]={"ch","bh","dh","ah","ERRORdi","ERRORsi","ERRORes","ERRORds"};
static char *lowword[]={"cx","bx","dx","ax","di","si","es","ds"};
static char *highword[]={"ERRORcx","cx","ERRORdx","dx","es","ds","ERRORes","ERRORds"};
#define NODEPTR_TYPE Node
#define OP_LABEL(p) ((p)->op)
#define LEFT_CHILD(p) ((p)->kids[0])
#define RIGHT_CHILD(p) ((p)->kids[1])
#define STATE_LABEL(p) ((p)->x.state)
static void address(Symbol, Symbol, int);
static void blkfetch(int, int, int, int);
static void blkloop(int, int, int, int, int, int[]);
static void blkstore(int, int, int, int);
static void defaddress(Symbol);
static void defconst(int, int, Value);
static void defstring(int, char *);
static void defsymbol(Symbol);
static void doarg(Node);
static void emit2(Node);
static void export(Symbol);
static void clobber(Node);
static void function(Symbol, Symbol [], Symbol [], int);
static void global(Symbol);
static void import(Symbol);
static void local(Symbol);
static void progbeg(int, char **);
static void progend(void);
static void printseg(Symbol);
static void segment(int);
static void space(int);
static void target(Node);
static int ckstack(Node, int);
static Symbol charreg[32], shortreg[32], lonreg[32], ptrreg[32];
static Symbol fltreg[32];
static long nbits_set(unsigned long j);
static int issmall(Node p);
static int is2n(Node p);
static Symbol charregw, shortregw, lonregw, ptrregw,fltregw;
 
char cseg_name[64];
static int cseg;
static short dsval;
static short esval;
  
%}
%start stmt

%term CNSTF4=4113
%term CNSTF8=8209
%term CNSTF16=16401

%term CNSTI1=1045
%term CNSTI2=2069
%term CNSTI4=4117
%term CNSTI8=8213

%term CNSTP2=2071

%term CNSTU1=1046
%term CNSTU2=2070
%term CNSTU4=4118
%term CNSTU8=8214


%term ARGB=41

%term ARGF4=4129
%term ARGF8=8225
%term ARGF16=16417

%term ARGI2=2085
%term ARGI4=4133
%term ARGI8=8229

%term ARGP2=2087

%term ARGU2=2086
%term ARGU4=4134
%term ARGU8=8230


%term ASGNB=57

%term ASGNF4=4145
%term ASGNF8=8241
%term ASGNF16=16433

%term ASGNI1=1077
%term ASGNI2=2101
%term ASGNI4=4149
%term ASGNI8=8245

%term ASGNP2=2103

%term ASGNU1=1078
%term ASGNU2=2102
%term ASGNU4=4150
%term ASGNU8=8246


%term INDIRB=73

%term INDIRF4=4161
%term INDIRF8=8257
%term INDIRF16=16449

%term INDIRI1=1093
%term INDIRI2=2117
%term INDIRI4=4165
%term INDIRI8=8261

%term INDIRP2=2119

%term INDIRU1=1094
%term INDIRU2=2118
%term INDIRU4=4166
%term INDIRU8=8262


%term CVFF4=4209
%term CVFF8=8305
%term CVFF16=16497

%term CVFI2=2165
%term CVFI4=4213
%term CVFI8=8309


%term CVIF4=4225
%term CVIF8=8321
%term CVIF16=16513

%term CVII1=1157
%term CVII2=2181
%term CVII4=4229
%term CVII8=8325

%term CVIU1=1158
%term CVIU2=2182
%term CVIU4=4230
%term CVIU8=8326


%term CVPU2=2198


%term CVUI1=1205
%term CVUI2=2229
%term CVUI4=4277
%term CVUI8=8373

%term CVUP2=2231

%term CVUU1=1206
%term CVUU2=2230
%term CVUU4=4278
%term CVUU8=8374


%term NEGF4=4289
%term NEGF8=8385
%term NEGF16=16577

%term NEGI2=2245
%term NEGI4=4293
%term NEGI8=8389


%term CALLB=217

%term CALLF4=4305
%term CALLF8=8401
%term CALLF16=16593

%term CALLI2=2261
%term CALLI4=4309
%term CALLI8=8405

%term CALLP2=2263

%term CALLU2=2262
%term CALLU4=4310
%term CALLU8=8406

%term CALLV=216


%term RETF4=4337
%term RETF8=8433
%term RETF16=16625

%term RETI2=2293
%term RETI4=4341
%term RETI8=8437

%term RETP2=2295

%term RETU2=2294
%term RETU4=4342
%term RETU8=8438

%term RETV=248


%term ADDRGP2=2311


%term ADDRFP2=2327


%term ADDRLP2=2343


%term ADDF4=4401
%term ADDF8=8497
%term ADDF16=16689

%term ADDI2=2357
%term ADDI4=4405
%term ADDI8=8501

%term ADDP2=2359

%term ADDU2=2358
%term ADDU4=4406
%term ADDU8=8502


%term SUBF4=4417
%term SUBF8=8513
%term SUBF16=16705

%term SUBI2=2373
%term SUBI4=4421
%term SUBI8=8517

%term SUBP2=2375

%term SUBU2=2374
%term SUBU4=4422
%term SUBU8=8518


%term LSHI2=2389
%term LSHI4=4437
%term LSHI8=8533

%term LSHU2=2390
%term LSHU4=4438
%term LSHU8=8534


%term MODI2=2405
%term MODI4=4453
%term MODI8=8549

%term MODU2=2406
%term MODU4=4454
%term MODU8=8550


%term RSHI2=2421
%term RSHI4=4469
%term RSHI8=8565

%term RSHU2=2422
%term RSHU4=4470
%term RSHU8=8566


%term BANDI2=2437
%term BANDI4=4485
%term BANDI8=8581

%term BANDU2=2438
%term BANDU4=4486
%term BANDU8=8582


%term BCOMI2=2453
%term BCOMI4=4501
%term BCOMI8=8597

%term BCOMU2=2454
%term BCOMU4=4502
%term BCOMU8=8598


%term BORI2=2469
%term BORI4=4517
%term BORI8=8613

%term BORU2=2470
%term BORU4=4518
%term BORU8=8614


%term BXORI2=2485
%term BXORI4=4533
%term BXORI8=8629

%term BXORU2=2486
%term BXORU4=4534
%term BXORU8=8630


%term DIVF4=4545
%term DIVF8=8641
%term DIVF16=16833

%term DIVI2=2501
%term DIVI4=4549
%term DIVI8=8645

%term DIVU2=2502
%term DIVU4=4550
%term DIVU8=8646


%term MULF4=4561
%term MULF8=8657
%term MULF16=16849

%term MULI2=2517
%term MULI4=4565
%term MULI8=8661

%term MULU2=2518
%term MULU4=4566
%term MULU8=8662


%term EQF4=4577
%term EQF8=8673
%term EQF16=16865

%term EQI2=2533
%term EQI4=4581
%term EQI8=8677

%term EQU2=2534
%term EQU4=4582
%term EQU8=8678


%term GEF4=4593
%term GEF8=8689
%term GEF16=16881

%term GEI2=2549
%term GEI4=4597
%term GEI8=8693

%term GEU2=2550
%term GEU4=4598
%term GEU8=8694


%term GTF4=4609
%term GTF8=8705
%term GTF16=16897

%term GTI2=2565
%term GTI4=4613
%term GTI8=8709

%term GTU2=2566
%term GTU4=4614
%term GTU8=8710


%term LEF4=4625
%term LEF8=8721
%term LEF16=16913

%term LEI2=2581
%term LEI4=4629
%term LEI8=8725

%term LEU2=2582
%term LEU4=4630
%term LEU8=8726


%term LTF4=4641
%term LTF8=8737
%term LTF16=16929

%term LTI2=2597
%term LTI4=4645
%term LTI8=8741

%term LTU2=2598
%term LTU4=4646
%term LTU8=8742


%term NEI2=2613
%term NEI4=4661
%term NEI8=8757

%term NEU2=2614
%term NEU4=4662
%term NEU8=8758


%term JUMPV=584


%term LABELV=600

%term LOADB=233
%term LOADF4=4321
%term LOADF8=8417
%term LOADF16=16609
%term LOADI1=1253
%term LOADI2=2277
%term LOADI4=4325
%term LOADI8=8421
%term LOADP2=2279
%term LOADU1=1254
%term LOADU2=2278
%term LOADU4=4326
%term LOADU8=8422

%term NEF4=4657
%term NEF8=8753

%term VREGP=711

%%

addr: local "%0" 1
addr: preg "%0" 1
addr: acon "%0" 1
addr: pcon "%0" 1

acon: ADDRGP2 "%a"
local: ADDRFP2 "bp+(%a)"
local: ADDRLP2 "bp+(%a)"
addr: ADDP2(preg,local) "%0+%1" 
addr: ADDP2(preg,i16) "%0+(%1)" 
addr: SUBP2(preg,i16) "%0-(%1)" 
addr: ADDP2(r16,i16) "%0+(%1)"
addr: ADDP2(addr,i16) "%0+(%1)" 
addr: SUBP2(addr,i16) "%0-(%1)"
addr: ADDP2(r16,preg) "%0+%1"
pcon: CNSTP2 "%a"
i16: pcon "%0" 1
i16: acon "%0" 1


preg: r32 "%0"
preg: i16 "mov %c,%0\n" 4
preg: addr "lea %c,[%0]\n" 4
r16: LOADU2(addr) "lea %c,[%0]\n" 2
zero:  CNSTP2 "%a" range(a, 0, 0)
preg:  INDIRI2(VREGP)     "# read register\n" 2
preg:  INDIRU2(VREGP)     "# read register\n" 2
preg: ADDP2(preg,minustwo) "?mov %c,%0\ndec %c\ndec %c\n" 2
preg: ADDP2(preg,minusone) "?mov %c,%0\ndec %c\n" 2
preg: ADDP2(preg,one) "?mov %c,%0\ninc %c\n" 2
preg: ADDP2(preg,two) "?mov %c,%0\ninc %c\ninc %c\n" 4
preg: ADDP2(preg,three) "?mov %c,%0\ninc %c\ninc %c\ninc %c\n" 6
r16: ADDP2(r16,rmi16) "?mov %c,%0\nadd %c,%1\n" 2
preg: ADDP2(preg,rmi16) "?mov %c,%0\nadd %c,%1\n" 2
preg: LOADP2(rmi16) "mov %c,%0\n" 2
preg: zero "xor %c,%c\n" 2
addr: ADDP2(acon,r16) "%0+%1" 1
addr: ADDP2(r16,ADDP2(r16,acon)) "%0+%1+%2" 1
addr: ADDP2(r16,acon) "%0+%1" 1
addr: ADDP2(preg,acon) "%0+%1" 1
addr: ADDP2(acon,preg) "%0+%1" 1
addr: ADDP2(r16,local) "%0+%1" 
addr: ADDP2(preg,local) "%0+%1" 
preg: INDIRP2(addr) "mov %c,word [%0]\n" 2
m16: INDIRP2(addr) "word [%0]" 
preg: ADDP2(preg,preg) "?mov %c,%0\nadd %c,%1\n" 2
preg: SUBP2(preg,rmi16) "?mov %c,%0\nadd %c,%1\n" 2
preg: CALLP2(acon) "call %0\nadd sp,%a\n" 6
preg: INDIRP2(VREGP)     "# read register\n" 2
preg: INDIRP2(mi16) "mov %c,%0\n" 4
preg: LOADP2(preg)  "# move\n"    2
r16: LOADP2(acon) "mov %c,%0\n" 4
r16: CALLI2(INDIRP2(addr)) "call [%0]\nadd sp,%a\n" 6
r16: CALLI2(preg) "call %0\nadd sp,%a\n" 6
r32: CALLI4(INDIRP2(addr)) "call [%0]\nadd sp,%a\n" 6
r32: CALLI4(preg) "call %0\nadd sp,%a\n" 6
r16: CALLU2(INDIRP2(addr)) "call [%0]\nadd sp,%a\n" 6
r16: CALLU2(preg) "call %0\nadd sp,%a\n" 6
r32: CALLU4(INDIRP2(addr)) "call [%0]\nadd sp,%a\n" 6
r32: CALLU4(preg) "call %0\nadd sp,%a\n" 6
freg: CALLF8(acon) "call %0\nadd sp, %a\n" 6
freg: CALLF8(preg) "call %0\nadd sp, %a\n" 6
freg: CALLF8(INDIRP2(addr)) "call [%0]\nadd sp,%a\n" 6
freg: CALLF4(acon) "call %0\nadd sp, %a\n" 6
freg: CALLF4(preg) "call %0\nadd sp, %a\n" 6
freg: CALLF4(INDIRP2(addr)) "call [%0]\nadd sp,%a\n" 6
stmt: ASGNI1(addr, LOADI1(ADDI2(CVII2(INDIRI2(ADDRGP2)),one))) "inc byte [%1]\n" 4
stmt: ASGNI1(addr, LOADI1(ADDI2(CVII2(INDIRI2(local)),one))) "inc byte [%1]\n" 3
stmt: ARGP2(preg) "push %0\n" 1
stmt: ARGP2(r16) "push %0\n" 1
stmt: ARGP2(INDIRP2(addr)) "push word [%0]\n" 2
stmt: ARGP2(r32)   "push hiword_%0\npush loword_%0\n" 2
stmt: ASGNI1(ADDRFP2,LOADI1(INDIRI2(ADDRFP2))) "# byte parameter\n" 0
stmt: ASGNU1(ADDRFP2,LOADU1(LOADU2(INDIRI2(ADDRFP2)))) "# byte parameter\n" 0
stmt: ASGNP2(VREGP,preg) "# write register\n" 2
stmt: ASGNP2(VREGP,r16) "# write register\n" 2
stmt: ASGNP2(addr, SUBP2(m16,one)) "dec word [%0]\n" 0
stmt: ASGNP2(addr, SUBP2(LOADP2(m16),one)) "dec word [%0]\n" 0
stmt: ASGNP2(addr, ADDP2(m16,one)) "inc word [%0]\n" 0
stmt: ASGNP2(addr, ADDP2(LOADP2(m16),one)) "inc word [%0]\n" 0
stmt: ASGNP2(addr, ADDP2(m16,minusone)) "dec word [%0]\n" 0
stmt: ASGNP2(addr, ADDP2(m16,two)) "inc word [%0]\ninc word [%0]\n" 2
stmt: ASGNP2(addr, ADDP2(m16,minustwo)) "dec word [%0]\ndec word [%0]\n" 2
stmt: ASGNP2(addr, SUBP2(m16,ri16)) "sub word [%0],%2\n" 0
stmt: ASGNP2(addr, ADDP2(m16,ri16)) "add word [%0],%2\n" 0
stmt: ASGNP2(addr, pcon) "mov word [%0],%1\n" 2
stmt: ASGNP2(addr, acon) "mov word [%0],%1\n" 2
stmt: ASGNP2(addr, preg) "mov word [%0],%1\n" 2
stmt: ASGNP2(addr, r16) "mov word [%0],%1\n" 2
stmt: JUMPV(INDIRP2(addr))  "jmp [%0]\n" 3
stmt: JUMPV(acon)  "jmp %0\n" 3
stmt: JUMPV(preg)  "jmp %0\n" 2
stmt: RETP2(preg)  "# ret\n" 2
r16: CALLU2(acon) "call %0\nadd sp,%a\n" 6 
r16: CALLI2(acon) "call %0\nadd sp,%a\n" 6 
r16: CALLI4(acon) "call %0\nadd sp,%a\n" 6 
r16: CALLU4(acon) "call %0\nadd sp,%a\n" 6 
r16: CALLV(acon) "call %0\nadd sp,%a\n" 6 
stmt: ARGB(INDIRB(addr)) "cli\nmov cx,%a\nsub sp,cx\nmov di,sp\nrep movsb\nsti\n" 11

stmt: LABELV "%a:\n"
freg:  INDIRF4(VREGP)     "# read register\n"
freg:  INDIRF8(VREGP)     "# read register\n"
freg: ADDF4(freg,freg) "call __fadd:0\n"
freg: ADDF8(freg,freg) "call __fadd:0\n"
freg: CVFF4(freg) "# CVFF4\n"
freg: CVFF8(freg) "# CVFF8\n"
freg: CVIF4(INDIRI2(preg)) "call __filds_%0:0\n"
freg: CVIF4(INDIRI4(preg)) "call __fildd_%0:0\n"
freg: CVIF4(r16)  "push %0\ncall __filds_sp:0\nadd sp,2\n"  6
freg: CVIF4(r32)  "push hiword_%0\npush loword_%0\ncall __fildd_sp:0\nadd sp,4\n"  6
freg: CVIF8(INDIRI2(preg)) "call __filds_%0:0\n"
freg: CVIF8(INDIRI4(preg)) "call __fildd_%0:0\n"
freg: CVIF8(r16)  "push %0\ncall __filds_sp:0\nadd sp,2\n"  6
freg: CVIF8(r32)  "push hiword_%0\npush loword_%0\ncall __fildd_sp:0\nadd sp,4\n"  6
freg: DIVF4(freg,freg) "call __fdiv:0\n"
freg: DIVF8(freg,freg) "call __fdiv:0\n"
freg: INDIRF4(preg) "call __fldl_%0:0\n"
freg: INDIRF8(preg) "call __fldd_%0:0\n"
freg: MULF4(freg,freg) "call __fmul:0\n"
freg: MULF8(freg,freg) "call __fmul:0\n"
freg: NEGF4(freg) "call __fchs:0\n"
freg: NEGF8(freg) "call __fchs:0\n"
freg: SUBF4(freg,freg) "call __fsub:0\n"
freg: SUBF8(freg,freg) "call __fsub:0\n"
r32: CVFI4(freg) "call __ftol:0\n"
r16: CVFI2(freg) "call __ftos:0\n"
stmt: ARGF4(freg) "sub sp,4\ncall __fstpl_sp:0\n"
stmt: ARGF8(INDIRF8(preg)) "cli\nsub sp,8\nmov di,sp\nmov cx,ss\nmov es,cx\nmov cx,4\nrep movsw\nsti\n"  (4+2+2+2+4+9+17*4)
stmt: ARGF8(freg) "sub sp,8\ncall __fstpd_sp:0\n"
stmt: ASGNF16(preg,INDIRF16(preg)) "mov cx,8\nrep movsw\n"  (4+9+17*8)
stmt: ASGNF4(VREGP,freg) "# write register\n"
stmt: ASGNF4(preg,INDIRF4(preg)) "mov cx,2\nrep movsw\n"  (4+9+17*2)
stmt: ASGNF4(preg,freg) "call __fstpl_%0:0\n"  
stmt: ASGNF8(VREGP,freg) "# write register\n"
stmt: ASGNF8(preg,INDIRF8(preg)) "mov cx,4\nrep movsw\n"  (4+9+17*4)
stmt: ASGNF8(preg,freg) "call __fstpd_%0:0\n"  
stmt: EQF4(freg,freg) "call __fcomp:0\nbe %a\n"
stmt: EQF8(freg,freg) "call __fcomp:0\nbe %a\n"
stmt: GTF4(freg,freg) "call __fcomp:0\nbg %a\n"
stmt: GEF4(freg,freg) "call __fcomp:0\nbge %a\n"
stmt: GEF8(freg,freg) "call __fcomp:0\nbge %a\n"
stmt: GTF8(freg,freg) "call __fcomp:0\nbg %a\n"
stmt: LEF4(freg,freg) "call __fcomp:0\nble %a\n"
stmt: LEF8(freg,freg) "call __fcomp:0\nble %a\n"
stmt: LTF4(freg,freg) "call __fcomp:0\nblt %a\n"
stmt: LTF8(freg,freg) "call __fcomp:0\nblt %a\n"
stmt: NEF4(freg,freg) "call __fcomp:0\nbne %a\n"
stmt: NEF8(freg,freg) "call __fcomp:0\nbne %a\n"
stmt: RETF4(freg)  "# ret\n"
stmt: RETF8(freg)  "# ret\n"
bit16: CNSTI2 "%a" range(a, -32768, 32767)
bit16: CNSTI4 "%a" range(a, -32768, 32767)
bit16: CNSTU2 "%a" range(a, 0, 65535)
bit16: CNSTU4 "%a" range(a, 0, 65535)

bit8: CNSTI1 "%a" range(a,-128,127)
bit8: CNSTI2 "%a" range(a,-128,127)
bit8: CNSTI4 "%a" range(a,-128,127)
bit8: CNSTU1 "%a" range(a,0,255)
bit8: CNSTU2 "%a" range(a,0,255)
bit8: CNSTU4 "%a" range(a,0,255)

con: icon "%0"
con: ucon "%0"

eight:  CNSTI1 "%a" range(a, 8, 8)
eight:  CNSTI2 "%a" range(a, 8, 8)
eight:  CNSTI4 "%a" range(a, 8, 8)
eight:  CNSTU1 "%a" range(a, 8, 8)
eight:  CNSTU2 "%a" range(a, 8, 8)
eight:  CNSTU4 "%a" range(a, 8, 8)

gt16:  CNSTI2 "%a" range(a, 17, 32767)
gt16:  CNSTI4 "%a" range(a, 17, 0x7FFFFFFF)
gt16:  CNSTU2 "%a" range(a, 17, 65535)
gt16:  CNSTU4 "%a" range(a, 17, 0xFFFFFFFF)

gt32:  CNSTI2 "%a" range(a, 33, 32767)
gt32:  CNSTI4 "%a" range(a, 33, 0x7FFFFFFF)
gt32:  CNSTU2 "%a" range(a, 33, 65535)
gt32:  CNSTU4 "%a" range(a, 33, 0xFFFFFFFF)

i16: bit16 "%0" 1

i8: bit8 "%0" 0

icon: CNSTI1  "%a"
icon: CNSTI2  "%a"
icon: CNSTI4  "%a"
icon: CNSTI8  "%a"

m16: INDIRI2(addr) "word [%0]" 
m16: INDIRU2(addr) "word [%0]" 

m8: INDIRI1(addr) "byte [%0]"  
m8: INDIRU1(addr) "byte [%0]"  

mi16: i16 "%0"
mi16: m16 "%0"

mi8: i8 "%0"
mi8: m8 "%0"

minusone:  CNSTI2 "%a" range(a, -1, -1)
minusone:  CNSTI4 "%a" range(a, -1, -1)

minustwo:  CNSTI2 "%a" range(a, -2, -2)
minustwo:  CNSTI4 "%a" range(a, -2, -2)

one:  CNSTI2 "%a" range(a, 1, 1)
one:  CNSTI4 "%a" range(a, 1, 1)
one:  CNSTU2 "%a" range(a, 1, 1)
one:  CNSTU4 "%a" range(a, 1, 1)

pow2: CNSTI2 "%a" is2n(a)
pow2: CNSTU2 "%a" is2n(a)
r16:  INDIRI2(VREGP)     "# read register\n" 2
r16:  INDIRU2(VREGP)     "# read register\n" 2
r16: ADDI2(r16,one) "?mov %c,%0\ninc %c\n" 1
r16: ADDI2(r16,rmi16) "?mov %c,%0\nadd %c,%1\n" 2
r16: ADDI2(r16,three) "?mov %c,%0\ninc %c\ninc %c\ninc %c\n" 3
r16: ADDI2(r16,two) "?mov %c,%0\ninc %c\ninc %c\n" 2
r16: ADDU2(r16,one) "?mov %c,%0\ninc %c\n" 1
r16: ADDU2(r16,rmi16) "?mov %c,%0\nadd %c,%1\n" 2
r16: ADDU2(r16,three) "?mov %c,%0\ninc %c\ninc %c\ninc %c" 3
r16: ADDU2(r16,two) "?mov %c,%0\ninc %c\ninc %c\n" 2
r16: BANDI2(r16,rmi16) "?mov %c,%0\nand %c,%1\n" 2
r16: BANDU2(r16,rmi16) "?mov %c,%0\nand %c,%1\n" 2
r16: BCOMI2(r16) "?mov %c,%0\nnot %c\n" 2
r16: BCOMU2(r16) "?mov %c,%0\nnot %c\n" 2
r16: BORI2(r16,rmi16) "?mov %c,%0\nor %c,%1\n" 2
r16: BORU2(r16,rmi16) "?mov %c,%0\nor %c,%1\n" 2
r16: BXORI2(r16,rmi16) "?mov %c,%0\nxor %c,%1\n" 2
r16: BXORU2(r16,rmi16) "?mov %c,%0\nxor %c,%1\n" 2
r16: CVII2(r8)  "cbw\n"  1
r16: CVUI2(r8)  "xor hibyte_%c,hibyte_%c\n" 2
r16: CVUU2(r8)  "xor hibyte_%c,hibyte_%c\n" 2
r16: DIVI2(r16,rm16) "?mov %c,%0\ncwd\nidiv %1\n" 3
r16: DIVI2(r16,two)  "?mov %c,%0\nsar %c,1\n"  2
r16: DIVU2(r16,rm16) "?mov %c,%0\nxor hiword_%c,hiword_%c\ndiv %1\n" 4
r16: DIVU2(r16,two)  "?mov %c,%0\nsar %c,1\n"  2
r16: LOADI2(r16)  "# move\n"   2
r16: LOADI2(r32)  "# move\n"   2
r16: LOADU2(r16)  "# move\n"   2
r16: LOADU2(r32)  "# move\n"   2
r16: LSHI2(r16,eight) "mov hibyte_%c,lobyte_%0\nxor lobyte_%c,lobyte_%c\n" 4
r16: LSHI2(r16,one) "?mov %c,%0\nsal %c,1\n" 2
r16: LSHI2(r16,r16) "?mov %c,%0\nsal %c,lobyte_%1\n" 2
r16: LSHI2(r16,two) "?mov %c,%0\nsal %c,1\nsal %c,1\n"  4
r16: LSHU2(r16,eight) "mov hibyte_%c,lobyte_%0\nxor lobyte_%c,lobyte_%c\n" 4
r16: LSHU2(r16,one) "?mov %c,%0\nshl %c,1\n" 2
r16: LSHU2(r16,r16) "?mov %c,%0\nshl %c,lobyte_%1\n" 2
r16: LSHU2(r16,two) "?mov %c,%0\nshl %c,1\nshl %c,1\n" 4
r16: MODI2(r16,pow2)  "?mov %c,%0\nand %c,%1-1\n" 3
r16: MODI2(r16,rm16) "cwd\nidiv %1\n" 3
r16: MODU2(r16,pow2)  "?mov %c,%0\nand %c,%1-1\n" 3
r16: MODU2(r16,rm16) "xor dx,dx\ndiv %1\n" 4
r16: MULI2(r16,rm16)  "?mov %c,%0\nimul %1\n" 2 
r16: MULI2(r16,two)   "?mov %c,%0\nsal %c,1\n" 2
r16: MULU2(r16,rm16)  "?mov %c,%0\nmul %1\n"  2
r16: MULU2(r16,two)   "?mov %c,%0\nsal %c,1\n" 2
r16: NEGI2(r16) "?mov %c,%0\nneg %c\n" 2
r16: RSHI2(INDIRI2(addr),eight) "mov lobyte_%c,byte [%0]\ncbw\n" 2
r16: RSHI2(INDIRI2(r16),gt16) "mov lobyte_%c,byte [%0+1]\ncbw\nmov lobyte_%c,hibyte_%c\n" 4
r16: RSHI2(r16,eight) "?mov %c,%0\nmov lobyte_%c,hibyte_%0\ncbw\n" 3
r16: RSHI2(r16,gt16) "mov lobyte_%c,hibyte_%0\ncbw\nmov lobyte_%c,hibyte_%c\n" 5
r16: RSHI2(r16,one) "?mov %c,%0\nsar %c,1\n" 2
r16: RSHI2(r16,r16) "?mov %c,%0\nsar %c,lobyte_%1\n" 2
r16: RSHI2(r16,two) "?mov %c,%0\nsar %c,1\nsar %c,1" 4
r16: RSHU2(INDIRU2(addr),eight) "mov lobyte_%c,byte [%0]\nxor hibyte_%c,hibyte_%c\n" 3
r16: RSHU2(r16,eight) "?mov %c,%0\nmov lobyte_%c,hibyte_%0\nxor hibyte_%c,hibyte_%c\n" 4
r16: RSHU2(r16,one) "?mov %c,%0\nshr %c,1\n" 2
r16: RSHU2(r16,r16) "?mov %c,%0\nshr %c,lobyte_%1\n" 2
r16: RSHU2(r16,two) "?mov %c,%0\nshr %c,1\nshr %c,1\n" 4
r16: SUBI2(r16,one) "?mov %c,%0\ndec %c\n" 2
r16: SUBI2(r16,rmi16) "?mov %c,%0\nsub %c,%1\n" 2
r16: SUBI2(r16,three) "?mov %c,%0\ndec %c\ndec %c\ndec %c\n" 2
r16: SUBI2(r16,two) "?mov %c,%0\ndec %c\ndec %c\n" 2
r16: SUBU2(r16,one) "?mov %c,%0\ndec %c\n" 2
r16: SUBU2(r16,rmi16) "?mov %c,%0\nsub %c,%1\n" 2
r16: SUBU2(r16,three) "?mov %c,%0\ndec %c\ndec %c\ndec %c\n" 2
r16: SUBU2(r16,two) "?mov %c,%0\ndec %c\ndec %c\n" 2
r16: i16 "mov %c,%0\n" 2
r16: m16 "mov %c,%0\n" 4
r16: zero "xor %c,%c\n" 2
r32:  INDIRI2(VREGP)     "xor hiword_%0,hiword_%0\ncmp loword_%0,0\nbge $+4\nneg hiword_%0\n" 9
r32:  INDIRI4(VREGP)     "# read register\n" 2
r32:  INDIRU2(VREGP)     "xor hiword_%0,hiword_%0\n" 2
r32:  INDIRU4(VREGP)     "# read register\n" 2
r32: ADDI4(r32,INDIRI4(addr)) "?movl_%c_%0\nadd loword_%c,word [%1]\nadc hiword_%c,word [%1+2]\n" 6
r32: ADDI4(r32,con) "?movl_%c_%0\nadd loword_%c,loword(%1)\nadc hiword_%c,hiword(%1)\n" 8
r32: ADDI4(r32,r32) "?movl_%c_%0\nadd loword_%c,loword_%1\nadc hiword_%c,hiword_%1\n" 4
r32: ADDU4(r32,INDIRU4(addr)) "?movl_%c_%0\nadd loword_%c,word [%1]\nadc hiword_%c,word [%1+2]\n" 6
r32: ADDU4(r32,con) "?movl_%c_%0\nadd loword_%c,loword(%1)\nadc hiword_%c,hiword(%1)\n" 8
r32: ADDU4(r32,r32) "?movl_%c_%0\nadd loword_%c,loword_%1\nadc hiword_%c,hiword_%1\n" 4
r32: BANDI4(r32,INDIRI4(addr)) "?movl_%c_%0\nand loword_%c,word [%1]\nand hiword_%c,word [%1+2]\n" 6
r32: BANDI4(r32,con) "?movl_%c_%0\nand hiword_%c,hiword(%1)\nand hiword_%c,hiword(%1)\n" 8
r32: BANDI4(r32,r32) "?movl_%c_%0\nand hiword_%c,hiword_%1\nand hiword_%c,hiword_%1\n" 4
r32: BANDU4(r32,con) "?movl_%c_%0\nand loword_%c,loword(%1)\nand hiword_%c,hiword(%1)\n" 8
r32: BCOMI2(r32) "?movl_%c_%0\nnot loword_%c\nnot hiword_%c\n" 4
r32: BCOMU2(r32) "?movl_%c_%0\nnot loword_%c\nnot hiword_%c\n" 4
r32: BORI4(r32,r32) "?movl_%c_%0\nor loword_%c,loword_%1\nor hiword_%c,hiword_%1\n" 4
r32: BXORI4(r32,r32) "?movl_%c_%0\nxor loword_%c,loword_%1\nxor hiword_%c,hiword_%1\n" 4
r32: CVII4(r16) "cwd\n"  1
r32: CVIU4(m8) "mov lobyte_%c,%0\nxor hibyte_%c,hibyte_%c\nxor hiword_%c,hiword_%c\n" 5
r32: CVIU4(mi16) "mov loword_%c,%0\nxor hiword_%c,hiword_%c\n" 3
r32: CVIU4(r16) "xor hiword_%c,hiword_%c\n" 2
r32: CVIU4(r8) "mov lobyte_%c,%0\nxor hibyte_%c,hibyte_%c\nxor hiword_%c,hiword_%c\n" 6
r32: CVUI4(r16) "xor hiword_%c,hiword_%c\n" 2
r32: CVUU4(r16) "xor hiword_%c,hiword_%c\n" 2
r32: DIVI4(r32,LOADI4(rm16)) "idiv %1\n" 2
r32: DIVI4(r32,r32) "call __divi4:0\n" 5
r32: DIVI4(r32,two)   "?movl_%c_%0\nsar hiword_%c,1\nrcr loword_%c,1\n" 4
r32: DIVU4(r32,LOADU4(rm16)) "div %1\n" 2
r32: DIVU4(r32,r32) "call __divu4:0\n" 5
r32: DIVU4(r32,two)   "?movl_%c_%0\nshr hiword_%c,1\nrcr loword_%c,1\n" 4
r32: INDIRI4(addr) "mov loword_%c,word [%0]\nmov hiword_%c,word [%0+2]\n" 4
r32: INDIRU4(addr) "mov loword_%c,word [%0]\nmov hiword_%c,word [%0+2]\n" 4
r32: LOADI4(pcon) "mov loword_%c,loword(%0)\nmov hiword_%c,hiword(%0)\n" 6
r32: LOADI4(r16) "# move\n" 2
r32: LOADI4(r32)  "# move\n"    2
r32: LOADU4(pcon) "mov loword_%c,loword(%0)\nmov hiword_%c,hiword(%0)\n" 6
r32: LOADU4(r32)  "# move\n"    2
r32: LSHI4(r32,one) "?movl_%c_%0\nshl loword_%c,1\nrcl hiword_%c,1\n" 4
r32: LSHI4(r32,r16) "?movl_%c_%0\nlshi4 %c,%1\n" 10
r32: LSHI4(r32,two) "?movl_%c_%0\nshl loword_%c,1\nrcl hiword_%c,1\nshl loword_%c,1\nrcl hiword_%c,1\n" 8
r32: LSHU4(r32,one) "?movl_%c_%0\nshl loword_%c,1\nrcl hiword_%c,1\n" 4
r32: LSHU4(r32,r16) "?movl_%c_%0\nlshu4 %c,%1\n" 10
r32: LSHU4(r32,two) "?movl_%c_%0\nshl loword_%c,1\nrcl hiword_%c,1\nshl loword_%c,1\nrcl hiword_%c,1\n" 8 
r32: MODI4(r32,r32) "call __modi4:0\n" 5
r32: MODU4(r32,r32) "call __modi4:0\n" 5
r32: MULI4(r32,r32)   "call __muli4:0\n" 5
r32: MULI4(r32,two)   "?movl_%c_%0\nshl loword_%c,1\nrcl hiword_%c,1\n" 4
r32: MULU4(r32,r32)   "call __mulu4:0\n" 5
r32: MULU4(r32,two)   "?movl_%c_%0\nshl loword_%c,1\nrcl hiword_%c,1\n" 4
r32: RSHI4(INDIRI4(addr),gt32) "mov loword_%c,word [%0+2]\ncwd\nmov hiword_%c,loword_%c\n" 4
r32: LSHI4(INDIRI4(addr),sixteen) "mov hiword_%c,word [%0]\nxor loword_%c,loword_%c\n" 2
r32: RSHI4(INDIRI4(addr),sixteen) "mov loword_%c,word [%0+2]\ncwd\n" 2
r32: RSHI4(r32,gt32) "mov loword_%c,hiword_%0\ncwd\nmov loword_%c,hiword_%c\n" 5
r32: RSHI4(r32,one) "?movl_%c_%0\nsar hiword_%c,1\nrcr loword_%c,1\n" 4
r32: RSHI4(r32,r16) "?movl_%c_%0\nrshi4 %c,%1\n" 10
r32: LSHI4(r32,sixteen) "mov hiword_%c,loword_%0\nxor loword_%0,loword_%0\n" 3
r32: RSHI4(r32,sixteen) "mov loword_%c,hiword_%0\ncwd\n" 3
r32: RSHI4(r32,two) "?movl_%c_%0\nsar hiword_%c,1\nrcr loword_%c,1\nsar hiword_%c,1\nrcr loword_%c,1\n" 8
r32: LSHU4(INDIRU4(addr),sixteen) "mov hiword_%c,word [%0]\nxor loword_%c,loword_%c\n" 3
r32: RSHU4(INDIRU4(addr),sixteen) "mov loword_%c,word [%0+2]\nxor hiword_%c,hiword_%c\n" 3
r32: RSHU4(r32,one) "?movl_%c_%0\nshr hiword_%c,1\nrcr loword_%c,1\n" 4
r32: RSHU4(r32,r16) "?movl_%c_%0\nrshu4 %c,%1\n" 10
r32: LSHU4(r32,sixteen) "mov hiword_%c,loword_%0\nxor loword_%c,loword_%c\n" 4
r32: RSHU4(r32,sixteen) "mov loword_%c,hiword_%0\nxor hiword_%c,hiword_%c\n" 4
r32: RSHU4(r32,two) "?movl_%c_%0\nshr hiword_%c,1\nrcr loword_%c,1\nsar hiword_$c,1\nrcr loword_%c,1\n" 8
r32: SUBI4(r32,INDIRI4(addr)) "?movl_%c_%0\nsub loword_%c,word [%1]\nsbb hiword_%c,word [%1+2]\n" 4
r32: SUBI4(r32,con) "?movl_%c_%0\nsub loword_%c,loword(%1)\nsbb hiword_%c,hiword(%1)\n" 6
r32: SUBI4(r32,r32) "?movl_%c_%0\nsub loword_%c,loword_%1\nsbb hiword_%c,hiword_%1\n" 4
r32: SUBU4(r32,INDIRU4(addr)) "?movl_%c_%0\nsub loword_%c,word [%1]\nsbb hiword_%c,word [%1+2]\n" 4
r32: SUBU4(r32,con) "?movl_%c_%0\nsub loword_%c,loword(%1)\nsbb hiword_%c,hiword(%1)\n" 6
r32: SUBU4(r32,r32) "?movl_%c_%0\nsub loword_%c,loword_%1\nsbb hiword_%c,hiword_%1\n" 4
r32: con "mov hiword_%c,hiword(%0)\nmov loword_%c,loword(%0)\n" 6
r32: zero "xor hiword_%c,hiword_%c\nxor loword_%c,loword_%c\n" 4
r8:  INDIRI1(VREGP)     "# read register\n" 2
r8:  INDIRU1(VREGP)     "# read register\n" 2
r8: ADDI2(r8,rmi8)  "?mov %c,%0\nadd lobyte_%c,lobyte_%1\n" 2
r8: ADDU2(r8,rmi8)  "?mov %c,%0\nadd lobyte_%c,lobyte_%1\n" 2
r8: LOADI1(r16)  "# move\n"    2
r8: LOADI1(r32)  "# move\n"    2
r8: LOADI1(r8)  "# move\n"    2
r8: LOADU1(r16)  "# move\n"    2
r8: LOADU1(r32)  "# move\n"    2
r8: LOADU1(r8)  "# move\n"   2
r8: mi8  "mov lobyte_%c,%0\n" 2
r8: zero "xor lobyte_%c,lobyte_%c\n"  2
ri16: i16 "%0"
ri16: r16 "%0"
ri8: i8 "%0"
ri8: r8 "lobyte_%0"
rm16: m16 "%0"
rm16: r16 "%0"
rm8: m8 "%0"
rm8: r8 "lobyte_%0"
rmi16: m16 "%0"
rmi16: ri16 "%0"
rmi8: i8 "%0"
rmi8: rm8 "%0"
fifteen:  CNSTI2 "%a" range(a, 15, 15)
fifteen:  CNSTU2 "%a" range(a, 15, 15)
fifteen:  CNSTI4 "%a" range(a, 15, 15)
fifteen:  CNSTU4 "%a" range(a, 15, 15)
sixteen:  CNSTI2 "%a" range(a, 16, 16)
sixteen:  CNSTI4 "%a" range(a, 16, 16)
sixteen:  CNSTU2 "%a" range(a, 16, 16)
sixteen:  CNSTU4 "%a" range(a, 16, 16)

r16: RSHI2(r16,fifteen) "?mov %c,%0\nadd %c,%c\nsbb %c,%c\n" 4
r16: RSHU2(r16,fifteen) "?mov %c,%0\nrol %c,1\nand %c,1\n" 4
r16: LSHI2(r16,fifteen) "?mov %c,%0\nror %c,1\nand %c,8000h\n" 4
r16: LSHU2(r16,fifteen) "?mov %c,%0\nror %c,1\nand %c,8000h\n" 4
stmt: ARGI2(rm16)  "push %0\n" 1
stmt: ARGI4(INDIRI4(addr)) "push word [%0+2]\npush word [%0]\n" 4
stmt: ARGI4(r32)   "push hiword_%0\npush loword_%0\n" 2
stmt: ARGU2(rm16)  "push %0\n" 1
stmt: ARGU4(INDIRU4(addr)) "push word [%0+2]\npush word [%0]\n" 4
stmt: ARGU4(r32)   "push hiword_%0\npush loword_%0\n" 2
stmt: ASGNB(addr,INDIRB(addr)) "mov cx,%a\nrep movsb\n" 5
stmt: ASGNF16(preg,INDIRF16(preg)) "mov cx,8\nrep movsw\n"  5
stmt: ASGNI1(VREGP,r8) "# write register\n" 2
stmt: ASGNI1(addr, LOADU1(ADDU2(CVUU2(m8),one))) "inc %1\n" 0
stmt: ASGNI1(addr, LOADU1(ADDU2(CVUU2(m8),two))) "inc %1\ninc %1\n" 2
stmt: ASGNI1(addr, LOADI1(ADDI2(CVII2(m8),one))) "inc %1\n" 0
stmt: ASGNI1(addr, LOADI1(ADDI2(CVII2(m8),two))) "inc %1\ninc %1\n" 2
stmt: ASGNI1(addr, ADDI2(m8,minusone)) "dec %1\n" 0
stmt: ASGNI1(addr, ADDI2(m8,minustwo)) "dec %1\ndec %1\n" 2
stmt: ASGNI1(addr, ADDI2(m8,ri8))  "add %1,%2\n" 0
stmt: ASGNI1(addr, BANDI2(m8,ri8)) "and %1,%2\n" 0
stmt: ASGNI1(addr, BORI2(m8,ri8)) "or %1,%2\n" 0
stmt: ASGNI1(addr, BXORI2(m8,ri8)) "xor %1,%2\n" 0
stmt: ASGNI1(addr, SUBI2(m8,one)) "dec %1\n" 0
stmt: ASGNI1(addr, SUBI2(m8,ri8))  "sub %1,%2\n" 0
stmt: ASGNI1(addr,ri8) "mov byte [%0],%1\n" 2
stmt: ASGNI1(rm8,r32) "mov %0,lobyte_%1\n" 2
stmt: ASGNI1(rm8,ri8) "mov %0,%1\n" 2
stmt: ASGNI2(VREGP,r16) "# write register\n" 2
stmt: ASGNI2(addr, ADDI2(m16,minusone)) "dec %1\n" 0
stmt: ASGNI2(addr, ADDI2(m16,minustwo)) "dec %1\ndec %1\n" 2
stmt: ASGNI2(addr, LSHI2(m16,one)) "sal %1,1\n" 0
stmt: ASGNI2(addr, RSHI2(m16,one)) "sar %1,1\n" 0
stmt: ASGNU2(addr, LSHU2(m16,one)) "shl %1,1\n" 0
stmt: ASGNU2(addr, RSHU2(m16,one)) "shr %1,1\n" 0
stmt: ASGNI2(addr, ADDI2(m16,one)) "inc %1\n" 0
stmt: ASGNI2(addr, ADDI2(m16,ri16))  "add %1,%2\n" 0
stmt: ASGNI2(addr, ADDI2(m16,three)) "inc %1\ninc %1\ninc %1\n" 8
stmt: ASGNI2(addr, ADDI2(m16,two)) "inc %1\ninc %1\n" 4
stmt: ASGNI2(addr, BANDI2(m16,ri16)) "and %1,%2\n" 0
stmt: ASGNI2(addr, BORI2(m16,ri16)) "or %1,%2\n" 0
stmt: ASGNI2(addr, BXORI2(m16,ri16)) "xor %1,%2\n" 0
stmt: ASGNI2(addr, NEGI2(m16)) "neg %1\n" 0
stmt: ASGNI2(addr, SUBI2(m16,one)) "dec %1\n" 0
stmt: ASGNI2(addr, SUBI2(m16,ri16))  "sub %1,%2\n" 4
stmt: ASGNI2(addr, SUBI2(m16,three)) "dec %1\ndec %1\ndec %1\n" 8
stmt: ASGNI2(addr, SUBI2(m16,two)) "dec %1\ndec %1\n" 4
stmt: ASGNI2(addr,BCOMI2(m16)) "not %1\n" 0
stmt: ASGNI2(addr,ri16) "mov word [%0],%1\n" 2
stmt: ASGNI4(VREGP,r32) "# write register\n" 2
stmt: ASGNI4(addr, BANDI4(INDIRI4(addr),con)) "and word [%1],loword(%2)\nand word [%1+2],hiword(%2)\n" 10
stmt: ASGNI4(addr, BANDI4(INDIRI4(addr),r32)) "and word [%1],loword_%2\nand word [%1+2],hiword_%2\n" 6
stmt: ASGNI4(addr, BORI4(INDIRI4(addr),con)) "or word [%1],loword(%2)\nor word [%1+2],hiword(%2)\n" 10
stmt: ASGNI4(addr, BORI4(INDIRI4(addr),r32)) "or word [%1],loword_%2\nor word [%1+2],hiword_%2\n" 6
stmt: ASGNI4(addr, BXORI4(INDIRI4(addr),con)) "xor word [%1],loword(%2)\nxor word [%1+2],hiword(%2)\n" 10
stmt: ASGNI4(addr, BXORI4(INDIRI4(addr),r32)) "xor word [%1],loword_%2\nxor word [%1+2],hiword_%2\n" 6
stmt: ASGNI4(addr,ADDI4(INDIRI4(addr),con)) "add word [%0],loword(%2)\nadc word [%0+2],hiword(%2)\n" 10
stmt: ASGNI4(addr,ADDI4(INDIRI4(addr),r32)) "add word [%0],loword_%2\nadc word [%0+2],hiword_%2\n" 6
stmt: ASGNI4(addr,SUBI4(INDIRI4(addr),con)) "sub word [%0],loword(%2)\nsbb word [%0+2],hiword(%2)\n" 10
stmt: ASGNI4(addr,SUBI4(INDIRI4(addr),r32)) "sub word [%0],loword_%2\nsbb word [%0+2],hiword_%2\n" 6
stmt: ASGNI4(addr,con) "mov word [%0],loword(%1)\nmov word [%0+2],hiword(%1)\n" 10
stmt: ASGNI4(addr,r32) "mov word [%0],loword_%1\nmov word [%0+2],hiword_%1\n" 6
stmt: ASGNU1(VREGP,r8) "# write register\n" 2
stmt: ASGNU1(addr, ADDU2(m8,one)) "inc %1\n" 0
stmt: ASGNU1(addr, ADDU2(m8,ri8))  "add %1,%2\n" 0
stmt: ASGNU1(addr, ADDU2(m8,three)) "inc %1\ninc %1\ninc %1\n" 8
stmt: ASGNU1(addr, ADDU2(m8,two)) "inc %1\ninc %1\n" 4
stmt: ASGNU1(addr, BANDU2(m8,ri8)) "and %1,%2\n" 0
stmt: ASGNU1(addr, BORU2(m8,ri8)) "or %1,%2\n" 0
stmt: ASGNU1(addr, SUBU2(m8,one)) "dec %1\n" 0
stmt: ASGNU1(addr, SUBU2(m8,ri8))  "add %1,%2\n" 0
stmt: ASGNU1(addr, SUBU2(m8,three)) "dec %1\ndec %1\ndec %1\n" 8
stmt: ASGNU1(addr, SUBU2(m8,two)) "dec %1\ndec %1\n" 4
stmt: ASGNU1(addr,ri8) "mov byte [%0],%1\n" 2
stmt: ASGNU2(VREGP,r16) "# write register\n" 2
stmt: ASGNU2(addr, ADDU2(m16,one)) "inc %1\n" 0
stmt: ASGNU2(addr, ADDU2(m16,ri16))  "add %1,%2\n" 0
stmt: ASGNU2(addr, ADDU2(m16,three)) "inc %1\ninc %1\ninc %1\n" 8
stmt: ASGNU2(addr, ADDU2(m16,two)) "inc %1\ninc %1\n" 4
stmt: ASGNU2(addr, BANDU2(m16,ri16)) "and %1,%2\n" 0
stmt: ASGNU2(addr, BORU2(m16,ri16)) "or %1,%2\n" 0
stmt: ASGNU2(addr, SUBU2(m16,one)) "dec %1\n" 0
stmt: ASGNU2(addr, SUBU2(m16,ri16))  "sub %1,%2\n" 0
stmt: ASGNU2(addr, SUBU2(m16,three)) "dec %1\ndec %1\ndec %1\n" 8
stmt: ASGNU2(addr, SUBU2(m16,two)) "dec %1\ndec %1\n" 4
stmt: ASGNU2(addr,BCOMU2(m16)) "not %1\n" 0
stmt: ASGNU2(addr,ri16) "mov word [%0],%1\n" 2
stmt: ASGNU4(VREGP,r32) "# write register\n" 2
stmt: ASGNU4(addr, BANDU4(INDIRU4(addr),con)) "and word [%1],loword(%2)\nand word [%1+2],hiword(%2)\n" 10
stmt: ASGNU4(addr, BANDU4(INDIRU4(addr),r32)) "and word [%1],loword_%2\nand word [%1+2],hiword_%2\n" 6
stmt: ASGNU4(addr, BORU4(INDIRU4(addr),con)) "or word [%1],loword(%2)\nor word [%1+2],hiword(%2)\n" 10
stmt: ASGNU4(addr, BORU4(INDIRU4(addr),r32)) "or word [%1],loword_%2\nor word [%1+2],hiword_%2\n" 6
stmt: ASGNU4(addr,ADDU4(INDIRU4(addr),con)) "add word [%0],loword(%2)\nadc word [%0+2],hiword(%2)\n" 10
stmt: ASGNU4(addr,ADDU4(INDIRU4(addr),r32)) "add word [%0],loword_%2\nadc word [%0+2],hiword_%2\n" 6
stmt: ASGNU4(addr,SUBU4(INDIRU4(addr),con)) "sub word [%0],loword(%2)\nsbb word [%0+2],hiword(%2)\n" 10
stmt: ASGNU4(addr,SUBU4(INDIRU4(addr),r32)) "sub word [%0],loword_%2\nsbb word [%0+2],hiword_%2\n" 6	
stmt: ASGNU4(addr,con) "mov word [%0],loword(%1)\nmov word [%0+2],hiword(%1)\n" 10
stmt: ASGNU4(addr,r32) "mov word [%0],loword_%1\nmov word [%0+2],hiword_%1\n" 6
stmt: EQI2(r16,zero) "or %0,%0\nbz %a\n" 7
stmt: EQI2(rm16,ri16) "cmp %0,%1\nbe %a\n" 7
stmt: EQI4(INDIRI4(addr),con) "eqi4_m32_con %0,%1,%a\n" 17
stmt: EQI4(INDIRI4(addr),r32) "eqi4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 13
stmt: EQI4(r32,con) "eqi4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: EQI4(r32,r32) "eqi4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 9
stmt: EQI4(r32,zero) "or hiword_%0,loword_%0\nbz %a\n" 7
stmt: EQU2(r16,zero) "or %0,%0\nbz %a\n" 7
stmt: EQU2(rm16,ri16) "cmp %0,%1\nbe %a\n" 7 
stmt: EQU4(INDIRU4(addr),con) "eqi4_m32_con %0,%1,%a\n" 17 
stmt: EQU4(INDIRU4(addr),r32) "eqi4_m32_con %0,loword_%1,hiword_%1,%a\n" 13
stmt: EQU4(r32,con) "eqi4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: EQU4(r32,r32) "eqi4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 9
stmt: EQU4(r32,zero) "or hiword_%0,loword_%0\nbz %a\n" 7
stmt: GEI2(CVII2(rm8),i8) "cmp %0,%1\nbge %a\n" 7
stmt: GEI2(rm16,ri16) "cmp %0,%1\nbge %a\n" 7
stmt: GEI4(CVII4(rm16),i16) "cmp %0,%1\nbge %a\n" 7
stmt: GEI4(INDIRI4(addr),con) "gei4_m32_con %0,%1,%a\n" 19
stmt: GEI4(INDIRI4(addr),r32) "gei4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: GEI4(r32,con) "gei4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: GEI4(r32,r32) "gei4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: GEI4(r32,zero) "or hiword_%0,hiword_%0\nbge %a\n" 7
stmt: GEU2(CVUU2(rm8),i8) "cmp %0,%1\nbae %a\n" 7
stmt: GEU2(rm16,ri16) "cmp %0,%1\nbae %a\n" 7
stmt: GEU4(CVUU4(rm16),i16) "cmp %0,%1\nbae %a\n" 7
stmt: GEU4(INDIRU4(addr),con) "geu4_m32_con %0,%1,%a\n" 19
stmt: GEU4(INDIRU4(addr),r32) "geu4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: GEU4(r32,con) "geu4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: GEU4(r32,r32) "geu4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: GTI2(CVII2(rm8),i8) "cmp %0,%1\nbg %a\n" 7
stmt: GTI2(rm16,ri16) "cmp %0,%1\nbg %a\n" 7
stmt: GTI4(CVII4(rm16),i16) "cmp %0,%1\nbg %a\n" 7
stmt: GTI4(INDIRI4(addr),con) "gti4_m32_con %0,%1,%a\n" 19
stmt: GTI4(INDIRI4(addr),r32) "gti4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: GTI4(r32,con) "gti4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: GTI4(r32,r32) "gti4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: GTU2(CVUU2(rm8),i8) "cmp %0,%1\nba %a\n" 7
stmt: GTU2(rm16,ri16) "cmp %0,%1\nba %a\n" 7
stmt: GTU4(CVUU4(rm16),i16) "cmp %0,%1\nba %a\n" 7
stmt: GTU4(INDIRU4(addr),con) "gtu4_m32_con %0,%1,%a\n" 19
stmt: GTU4(INDIRU4(addr),r32) "gtu4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: GTU4(r32,con) "gtu4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: GTU4(r32,r32) "gtu4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: JUMPV(acon)  "jmp %0\n" 3
stmt: LEI2(CVII2(rm8),i8) "cmp %0,%1\nble %a\n" 7
stmt: LEI2(rm16,ri16) "cmp %0,%1\nble %a\n" 7
stmt: LEI4(CVII4(rm16),i16) "cmp %0,%1\nble %a\n" 7
stmt: LEI4(INDIRI4(addr),con) "lei4_m32_con %0,%1,%a\n" 19
stmt: LEI4(INDIRI4(addr),r32) "lei4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: LEI4(r32,con) "lei4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: LEI4(r32,r32) "lei4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: LEU2(CVUU2(rm8),i8) "cmp %0,%1\nbbe %a\n" 7
stmt: LEU2(rm16,ri16) "cmp %0,%1\nbbe %a\n" 7
stmt: LEU4(CVUU4(rm16),i16) "cmp %0,%1\nbbe %a\n" 7
stmt: LEU4(INDIRU4(addr),con) "leu4_m32_con %0,%1,%a\n" 19
stmt: LEU4(INDIRU4(addr),r32) "leu4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: LEU4(r32,con) "leu4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: LEU4(r32,r32) "leu4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: LTI2(CVII2(rm8),i8) "cmp %0,%1\nblt %a\n" 7
stmt: LTI2(rm16,ri16) "cmp %0,%1\nblt %a\n" 7
stmt: LTI4(CVII4(rm16),i16) "cmp %0,%1\nblt %a\n" 7
stmt: LTI4(INDIRI4(addr),con) "lti4_m32_con %0,%1,%a\n" 19
stmt: LTI4(INDIRI4(addr),r32) "lti4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: LTI4(r32,con) "lti4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: LTI4(r32,r32) "lti4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: LTI4(r32,zero) "or hiword_%0,hiword_%0\nblt %a\n" 7
stmt: LTU2(CVUU2(rm8),i8) "cmp %0,%1\nbb %a\n" 7
stmt: LTU2(rm16,ri16) "cmp %0,%1\nbb %a\n" 7
stmt: LTU4(CVIU4(rm16),ri16) "cmp %0,%1\nbb %a\n" 7
stmt: LTU4(CVUU4(rm16),i16) "cmp %0,%1\nbb %a\n" 7
stmt: LTU4(INDIRU4(addr),con) "ltu4_m32_con %0,%1,%a\n" 19
stmt: LTU4(INDIRU4(addr),r32) "ltu4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 15
stmt: LTU4(r32,con) "ltu4_r32_con loword_%0,hiword_%0,%1,%a\n" 12
stmt: LTU4(r32,r32) "ltu4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 11
stmt: NEI2(CVII2(rm8),i8) "cmp %0,%1\nbne %a\n" 7
stmt: NEI2(r16,zero) "or %0,%0\nbz %a\n" 7
stmt: NEI2(rm16,ri16) "cmp %0,%1\nbne %a\n" 7
stmt: NEI4(CVII4(rm16),i16) "cmp %0,%1\nbne %a\n" 7
stmt: NEI4(INDIRI4(addr),con) "nei4_m32_con %0,%1,%a\n" 17
stmt: NEI4(INDIRI4(addr),r32) "nei4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 13
stmt: NEI4(r32,con) "nei4_r32_con loword_%0,hiword_%0,%1,%a\n" 10
stmt: NEI4(r32,r32) "nei4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 9
stmt: NEI4(r32,zero) "or hiword_%0,loword_%0\nbnz %a\n" 7
stmt: NEU2(CVUU2(rm8),i8) "cmp %0,%1\nbne %a\n" 7
stmt: NEU2(r16,rmi16) "cmp %0,%1\nbne %a\n" 7
stmt: NEU2(r16,zero) "or %0,%0\nbnz %a\n" 7
stmt: NEU2(rm16,ri16) "cmp %0,%1\nbne %a\n" 7
stmt: NEU4(CVUU4(rm16),i16) "cmp %0,%1\nbne %a\n" 7
stmt: NEU4(INDIRU4(addr),con) "nei4_m32_con %0,%1,%a\n" 17
stmt: NEU4(INDIRU4(addr),r32) "nei4_m32_r32 %0,loword_%1,hiword_%1,%a\n" 13
stmt: NEU4(r32,con) "nei4_r32_con loword_%0,hiword_%0,%1,%a\n" 10
stmt: NEU4(r32,r32) "nei4_r32_r32 loword_%0,hiword_%0,loword_%1,hiword_%1,%a\n" 9 
stmt: NEU4(r32,zero) "or hiword_%0,loword_%0\nbnz %a\n" 7
stmt: RETI2(r16)  "# ret\n"
stmt: RETI4(r32)  "# ret\n"
stmt: RETU2(r16)  "# ret\n"
stmt: RETU4(r32)  "# ret\n"
stmt: preg ""
stmt: r16 ""
stmt: r32 ""
stmt: r8 ""
three:  CNSTI2 "%a" range(a, 3, 3)
three:  CNSTI4 "%a" range(a, 3, 3)
three:  CNSTU2 "%a" range(a, 3, 3)
three:  CNSTU4 "%a" range(a, 3, 3)
two:  CNSTI2 "%a" range(a, 2, 2)
two:  CNSTI4 "%a" range(a, 2, 2)
two:  CNSTU2 "%a" range(a, 2, 2)
two:  CNSTU4 "%a" range(a, 2, 2)
ucon: CNSTU1  "%a"
ucon: CNSTU2  "%a"
ucon: CNSTU4  "%a"
ucon: CNSTU8  "%a"
zero: BANDI2(rm16,zero) "0"
zero: BANDI4(INDIRI4(addr),zero) "0"
zero: BANDI4(r32,zero) "0"
zero: BANDU2(rm16,zero) "0"
zero: BANDU4(INDIRU4(addr),zero) "0"
zero: BANDU4(r32,zero) "0"
zero: CNSTI2 "%a" range(a, 0, 0)
zero: CNSTI4 "%a" range(a, 0, 0)
zero: CNSTU2 "%a" range(a, 0, 0)
zero: CNSTU4 "%a" range(a, 0, 0)
zero: LSHI2(rm16,gt16) "0" 
zero: LSHI4(INDIRI4(addr),gt32) "0" 
zero: LSHI4(r32,gt32) "0" 
zero: LSHU2(rm16,gt16) "0" 
zero: LSHU4(INDIRU4(addr),gt32) "0"
zero: LSHU4(r32,gt32) "0"
zero: DIVI2(zero,rm16) "0"
zero: DIVU2(zero,rm16) "0"
zero: MULI2(zero,rm16) "0"
zero: MULU2(zero,rm16) "0"
zero: MULI2(rm16,zero) "0"
zero: MULU2(rm16,zero) "0"
zero: RSHU2(rm16,gt16) "0" 
zero: RSHU4(INDIRU4(addr),gt32) "0"
zero: RSHU4(r32,gt32) "0"
%%
#define ck(x) return ((x)?0:LBURG_MAX)

long nbits_set(unsigned long j) {
  int retval=0,i;
  for (i=0;i<(CHAR_BIT*sizeof(long));i++) {
     retval+=!!(j & (1<<i));
  }
  return(retval);
}

int issmall(Node p) {
  Symbol s = p->syms[0];
  switch (specific(p->op)) {
     case CNST+I: return (nbits_set(s->u.c.v.i)*(15+8));
     case CNST+U: return (nbits_set(s->u.c.v.u)*(15+8));
     case CNST+P: return (nbits_set(s->u.c.v.u)*(15+8)*2);
  }
  return LBURG_MAX;
}


int is2n(Node p) {
	Symbol s = p->syms[0];
	switch (specific(p->op)) {
	  case CNST+I: ck(nbits_set(s->u.c.v.i)==1);
	  case CNST+U: ck(nbits_set(s->u.c.v.u)==1);
	  case CNST+P: ck(nbits_set((unsigned long)(s->u.c.v.p))==1);
	}
	return LBURG_MAX;
}

static void progbeg(int argc, char *argv[]) {
        int i;

        {
                union {
                        char c;
                        int i;
                } u;
                u.i = 0;
                u.c = 1;
                swap = ((int)(u.i == 1)) != IR->little_endian;
        }
        parseflags(argc, argv);

	lonreg[AX] = mkreg("ax", AX, 1, IREG);
	lonreg[AX]->x.regnode->mask |= ((1<<AX) | (1<<DX));
	lonreg[BX] = mkreg("bx", BX, 1, IREG);
	lonreg[BX]->x.regnode->mask |= ((1<<BX) | (1<<CX));

	ptrreg[SI] = mkreg("si", SI, 1, IREG);
	ptrreg[DI] = mkreg("di", DI, 1, IREG);

        shortreg[AX] = mkreg("ax", AX, 1, IREG);
        shortreg[CX] = mkreg("cx", CX, 1, IREG);
        shortreg[DX] = mkreg("dx", DX, 1, IREG);
        shortreg[BX] = mkreg("bx", BX, 1, IREG);
  	ptrreg[BX] = shortreg[BX]; 

        charreg[AX]  = mkreg("ax", AX, 1, IREG);
        charreg[CX]  = mkreg("cx", CX, 1, IREG);
        charreg[DX]  = mkreg("dx", DX, 1, IREG);
        charreg[BX]  = mkreg("bx", BX, 1, IREG);

	for (i=0;i<8;i++) fltreg[i]=mkreg("%d",i,0,FREG);

        charregw = mkwildcard(charreg);
        shortregw = mkwildcard(shortreg);
	ptrregw = mkwildcard(ptrreg);
        lonregw = mkwildcard(lonreg);
	fltregw = mkwildcard(fltreg);

        tmask[IREG] = (1<<AX) | (1<<BX) | (1<<CX) | (1<<DX)
		    | (1<<SI) | (1<<DI) | (1<<ES) | (1<<DS);
        vmask[IREG] = 0;
        tmask[FREG] = 0xff;
        vmask[FREG] = 0;
	print("%%include %csupport.inc%c\n",34,34);
        print("BITS 16\n");
        cseg=0;
}

static Symbol rmap(int opk) {
        switch (optype(opk)) {
        case B: case P:
                return ptrregw;
        case I: case U:
                if (opsize(opk) == 1) {
                        return charregw;
		}
                else if (opsize(opk) == 2) {
                        return shortregw;
		}
                else {
                        return lonregw;
		}
	case F:
		return fltregw;
        default:
                return 0;
        }
}

static void segment(int n) {
        char *class;
        if (cseg!=n) {
	  cseg=n;
  	  switch (cseg) {
          case DATA: case LIT:
		     class=".data";
		     break;
	  case BSS: 
		     class=".bss";
		     break;
          default:   class=".text";
		     break;
	  }
	  print("SECTION %s\n",class);
	}
}

static void progend(void) {
        cseg=0;
        print("end\n");
}

static void target(Node p) {
        assert(p);
        switch (specific(p->op)) {
/*  	case LOAD+P:
		fprintf(stderr,"Targeting!!!");
		fprintf(stderr," generic(p->kids[0]->op)=%d\n",generic(p->kids[0]->op));
		if (generic(p->kids[0]->op)==VREG) {
		  if ((p->kids[0]->syms[RX]==lonregw)    ||
		      (p->kids[0]->syms[RX]==lonreg[AX]) ||
		      (p->kids[0]->syms[RX]==lonreg[BX])) {
		        setreg(p, ptrregw);
		  }
		}
		break;
*/
/*        case LOAD+U:
		if (specific(p->kids[0]->op)==VREG+P) {
		  setreg(p, lonreg[AX]);
		}
		break;
*/
        case RSH+I: case RSH+U: case LSH+I: case LSH+U:
		debug(fprintf(stderr,"Targeting for RSH/LSH\n"));
                if ((generic(p->kids[1]->op) != CNST)
		  || (range(p->kids[1],1,2)!=0))
		{
                        rtarget(p, 1, charreg[CX]);
		        switch (opsize(p->op)) {
		        case 4: rtarget(p,0,lonreg[AX]);
		          setreg(p,lonreg[AX]);
			  break;
 /*			case 2: rtarget(p,0,shortreg[AX]);
			  setreg(p,shortreg[AX]);
			  break;
 */
                        }
                }
                break;
        case MUL+U: case MUL+I:
		debug(fprintf(stderr,"Targeting for MUL\n"));
                if (opsize(p->op) != 4) {
                  setreg(p, shortreg[AX]);
                  rtarget(p, 0, shortreg[AX]);
                } else {
                  setreg(p, lonreg[AX]);
                  rtarget(p, 0, lonreg[AX]);
                } 
                break;
        case DIV+I: case DIV+U:
		debug(fprintf(stderr,"Targeting for DIV\n"));
		switch (opsize(p->op)) {
		  case 4: setreg(p, lonreg[AX]);
                          rtarget(p, 0, lonreg[AX]);
			  break;
		  case 2: setreg(p, shortreg[AX]);
                          rtarget(p, 0, shortreg[AX]);
			  break;
		}
                break;
	case SUB+I: case SUB+U:
		switch (opsize(p->op)) {
		   case 4: rtarget(p,0,lonreg[AX]);
		           setreg(p,lonreg[AX]);
			   break;
/*		   case 2: rtarget(p,0,shortreg[AX]);
		           setreg(p,shortreg[AX]);
			   break;
                   case 1: rtarget(p,0,charreg[AX]);
			   setreg(p,charreg[AX]);
			   break;
*/
		}
		break;
        case ADD+P: case SUB+P:
		debug(fprintf(stderr,"Targeting for ADDP\n"));
		if ((generic(p->kids[1]->op) != ADD) &&
		    (generic(p->kids[1]->op) != ADDRL) &&
		    (generic(p->kids[1]->op) != ADDRF))
		{
             	  rtarget(p, 0, ptrreg[BX]);  
		} else {
		  rtarget(p, 0, ptrreg[DI]);
		}
           	if ((generic(p->kids[1]->op) != CNST) &&
		    (generic(p->kids[1]->op) != ADDRG) &&
		    (generic(p->kids[1]->op) != ADDRL)) {
		  rtarget(p, 0, ptrreg[SI]);
		}
                break;
        case MOD+I: case MOD+U:
		debug(fprintf(stderr,"Targeting for MOD\n"));
                if (opsize(p->op) != 4) {
		  setreg(p, shortreg[DX]);
		  rtarget(p, 0, shortreg[AX]);
		  rtarget(p, 1, shortreg[BX]);
		} else {
		  setreg(p, lonreg[AX]);
                  rtarget(p, 0, lonreg[AX]);
                  rtarget(p, 1, lonreg[BX]);
		}
                break;
	case ASGN+B: case ASGN+F:
		debug(fprintf(stderr,"Targeting for ASGN+B|F\n"));
		rtarget(p, 0, ptrreg[SI]);
		if (specific(p->kids[1]->op)!=CVI+F)
		  rtarget(p, 1, ptrreg[DI]);
		break;
        case ARG+B: case ARG+F:
		debug(fprintf(stderr,"Targeting for ARG+B|F\n"));
                rtarget(p, 0, ptrreg[SI]);
                break;
        case CVF+I:
		debug(fprintf(stderr,"Targeting for CVFI\n"));
                rtarget(p, 0, lonreg[AX]);
                break;
	case CVU+I: case CVI+I: 
		debug(fprintf(stderr,"Targeting for CVU|I+I\n"));
		rtarget(p, 0, lonreg[AX]);
		setreg(p, lonreg[AX]);
		break;
        case CVP+U:
		debug(fprintf(stderr,"Targeting for CVPU\n"));
		rtarget(p,0,lonreg[AX]);
                break;
	case CVU+P:
		debug(fprintf(stderr,"Targeting for CVUP\n"));
		setreg(p, lonreg[AX]);
		break;
        case CALL+I: case CALL+U: case CALL+V:
		debug(fprintf(stderr,"Targeting for CALL+I|U|V\n"));
                setreg(p, lonreg[AX]);
                break;
	case CALL+P: 
		debug(fprintf(stderr,"Targeting for CALLP\n"));
		setreg(p, ptrreg[DI]);
		break;
        case RET+I: case RET+U: 
		debug(fprintf(stderr,"Targeting for RET+I|U\n"));
                rtarget(p, 0, lonreg[AX]);
                break;
	case RET+P: 
		debug(fprintf(stderr,"Targeting for RET+P\n"));
                rtarget(p, 0, ptrreg[DI]);
                break;
        }
	fflush(stdout);
}

static void clobber(Node p) {
        static int nstack = 0;

        assert(p);
        nstack = ckstack(p, nstack);
        switch (specific(p->op)) {
        case ASGN+B: case ARG+B: case ASGN+F: case ARG+F:
		debug(fprintf(stderr,"clobbering for ASGN|ARG+B|F\n"));
                spill(1<<CX | 1<<SI | 1<<DI, IREG, p);
                break;
	case MUL+U: case MUL+I:
		debug(fprintf(stderr,"clobbering for MUL+U|I\n"));
		switch (opsize(p->op)) {
		  case 2: spill(1<<DX,IREG,p); 
			  break;
		  case 4: break;
		}
		break;
        case DIV+U: case DIV+I: 
		debug(fprintf(stderr,"clobbering for DIV+U|I\n"));
		if (opsize(p->op)==4) {
		  spill(1<<CX | 1<<BX, IREG, p);
		} else {
		  spill(1<<DX, IREG, p);
		}
		break;
        case MOD+U: case MOD+I:
		debug(fprintf(stderr,"clobbering for MOD+U|I\n"));
		if (opsize(p->op)!=4) {
		  spill(1<<AX, IREG, p);
		}
		break;
	case EQ+U: case EQ+I:
		debug(fprintf(stderr,"clobbering for EQ+U|I\n"));
		if ((opsize(p->op)==4) && (generic(p->kids[1]->op) == CNST))
		  spill(1<<CX | 1<<DX, IREG, p);
		break;
        case CALL+U: case CALL+I: 
		debug(fprintf(stderr,"clobbering for CALL+U|I\n"));
		if (opsize(p->op)==4) {
		  spill(1<<BX | 1<<CX | 1<<SI | 1<<DI , IREG, p);
		} else {
		  spill(1<<BX | 1<<CX | 1<<DX | 1<<SI | 1<<DI , IREG, p);
		}
		break;
	case CALL+P: case CALL+B: case CALL+F:
		debug(fprintf(stderr,"clobbering for CALL+P|B|F\n"));
		spill(1<<AX | 1<<DX | 1<<BX | 1<<CX | 1<<SI , IREG, p);
		break;
	}
	fflush(stdout);
}

#define isfp(p) (optype((p)->op)==F)

static int ckstack(Node p, int n) {
        int i;

        for (i = 0; i < NELEMS(p->x.kids) && p->x.kids[i]; i++)
                if (isfp(p->x.kids[i]))
                        n--;
        if (isfp(p) && p->count > 0)
                n++;
        if (n > 8)
                error("expression too complicated\n");
        assert(n >= 0);
        return n;
}

static void emit2(Node p) {
        int op = specific(p->op);
#define preg(f) ((f)[getregnum(p->x.kids[0])]->x.name)
        debug(fprintf(stderr,"emit2: specific op = %d\n",op));
	fflush(stderr);
	fflush(stdout);
        if (generic(op) == CVI || generic(op) == CVU || generic(op) == LOAD) {
		int dstreg=getregnum(p);
                int srcreg=getregnum(p->kids[0]);
                assert(opsize(p->op) <= opsize(p->x.kids[0]->op));
                if (dstreg != srcreg) {
		   print("; Is this a necessary move?\n");
		   switch (opsize(p->op)) {
		     case 1: print("mov %s,%s\n",charreg[dstreg]->x.name,charreg[srcreg]->x.name);
			     break;
		     case 2: print("mov %s,%s\n",lowword[dstreg],lowword[srcreg]);
			     break;
		     case 4: if (optype(p->op)!=P) {
			       print("mov %s,%s\nmov %s,%s\n",
				      lowword[dstreg],lowword[srcreg],
				      highword[dstreg],highword[srcreg]);
			     } else {
			       print("cli\nmov %s,%s\nmov %s,%s\nmov %s,%s\nsti\n",
				      lowword[dstreg],highword[srcreg],
				      highword[dstreg],lowword[dstreg],
				      lowword[dstreg],lowword[srcreg]);
			     }
			     break;
		    }
		}
        }
	fflush(stdout);
}

static void doarg(Node p) {
        assert(p && p->syms[0]);
        mkactual(2, p->syms[0]->u.c.v.i);
}

static void blkfetch(int k, int off, int reg, int tmp) {}
static void blkstore(int k, int off, int reg, int tmp) {}
static void blkloop(int dreg, int doff, int sreg, int soff,
        int size, int tmps[]) {}

static void local(Symbol p) {
        if (isfloat(p->type))
                p->sclass = AUTO;
        if (askregvar(p, rmap(ttob(p->type))) == 0)
                mkauto(p);
}

static void function(Symbol f, Symbol caller[], Symbol callee[], int n) {
        int i;
	printseg(f);
	print("%s:\n",f->x.name);
        print("push bp\n");
        print("mov bp,sp\n");
        usedmask[0] = usedmask[1] = 0;
        freemask[0] = freemask[1] = ~(unsigned)0;
        offset = 2 + 2;
        for (i = 0; callee[i]; i++) {
                Symbol p = callee[i];
                Symbol q = caller[i];
                assert(q);
                p->x.offset = q->x.offset = offset;
                p->x.name = q->x.name = stringf("%d", p->x.offset);
                p->sclass = q->sclass = AUTO;
                offset += roundup(q->type->size, 2);
        }
        assert(caller[i] == 0);
        offset = maxoffset = 0;
        gencode(caller, callee);
        framesize = roundup(maxoffset, 2);
        if (framesize > 0)
                print("sub sp,%d\n", framesize);
        fflush(stdout);
        emitcode();
        print("mov sp,bp\n");
        print("pop bp\n");
        print("ret\n");
	print("END%s:\n",f->x.name);
	segment(DATA);
	print("GLOBAL SIZE%s\nSIZE%s:\ndd END%s-%s\n",
	       f->x.name,f->x.name,f->x.name,f->x.name,f->x.name);
}

void printseg(Symbol p) {
	segment(p->u.seg);
}

static void defsymbol(Symbol p) {
        int newseg=0;
        if (p->scope >= LOCAL && p->sclass == STATIC)
	{
                p->x.name = stringf("L%d", genlabel(1));
		newseg=1;
	}
        else if (p->generated)
	{
                p->x.name = stringf("L%s", p->name);
		newseg=1;
	}
        else if (p->scope == GLOBAL || p->sclass == EXTERN)
	{
                p->x.name = stringf("_%s", p->name);
		newseg=1;
	}
        else if (p->scope == CONSTANTS
        && (isint(p->type) || isptr(p->type))
        && p->name[0] == '0' && p->name[1] == 'x')
                p->x.name = stringf("0%sH", &p->name[2]);
        else
                p->x.name = p->name;
}

static void address(Symbol q, Symbol p, int n) {
        if (p->scope == GLOBAL
        || p->sclass == STATIC || p->sclass == EXTERN)
                q->x.name = stringf("%s%s%d",
                        p->x.name, n >= 0 ? "+" : "", n);
        else {
                q->x.offset = p->x.offset + n;
                q->x.name = stringd(q->x.offset);
        }
}
static void defconst(int suffix, int size, Value v) {
        if (suffix == I && size == 1)
                print("db %d\n",   v.u);
        else if (suffix == I && size == 2)
                print("dw %d\n",   v.i);
        else if (suffix == I && size == 4)
                print("dd %d\n",   v.i);
        else if (suffix == U && size == 1)
                print("db 0%xH\n", v.u);
        else if (suffix == U && size == 2)
                print("dw 0%xH\n", v.u);
        else if (suffix == U && size == 4)
                print("dd 0%xH\n", v.u);
        else if (suffix == P && size == 2)
                print("dw 0%xH\n", v.p);
        else if (suffix == F && size == 4) {
                float f = v.d;
                print("dd 0%xH\n",*(unsigned *)&f);
        }
        else if (suffix == F && size == 8) {
                double d = v.d;
		unsigned *p=(unsigned *)&d;
                print("dd 0%xH\ndd 0%xH\n", p[swap],p[!swap]);
        }
        else assert(0);
}
static void defaddress(Symbol p) {
        print("dw %s\n", p->x.name,p->x.name);
}

static void defstring(int n, char *str) {
        char *s;
	int quote=0;

        print("db ");
	for (s=str; s<str+n; s++) {
	  if (isprint(*s) && (*s != '\"')) {
	    if (!quote) {
	      quote=1;
	      print("\"");
	    }
	    print("%c",*s);
	  } else {
	    if (quote) {
	      quote=0;
	      print("\",");
	    }
	    print("%d",*s);
	    if (s<str+n-1) print(",");
	  }
	}
	if (quote) print("\"");
        print("\n");
}


static void export(Symbol p) {
        print("GLOBAL %s\n", p->x.name);
}

static void import(Symbol p) {
        if (p->ref > 0) {
	       print("EXTERN %s\n",p->x.name);
        }
}

static void global(Symbol p) {
        static char *rsize[]={"","db 0","dw 0","times 3 db (0)","dd 0","times 5 db (0)",
		       "times 3 dw (0)","times 7 db (0)","times 2 dd 0"};
        printseg(p);
        print("%s:\n", p->x.name);
/*        if (p->u.seg == DATA || p->u.seg == LIT) {
	  if (p->type->size < 9) 
                print("%s\n", rsize[p->type->size]);
	  else
		print("times %d db (0)\n", p->type->size);
        } 
 */
}

static void space(int n) {
        switch (cseg) {
	  case BSS: print("RESB %d\n", n);
		    break;
          case DATA: print("times %d db (0)\n",n);
                    break;
          default:  break;
	} 
}

Interface smallIR = {
        1, 1, 0,  /* char */
        2, 2, 0,  /* short */
        2, 2, 0,  /* int */
        4, 2, 0,  /* long */
        4, 2, 1,  /* float */
        8, 2, 1,  /* double */
        8, 2, 1,  /* long double */
        2, 2, 0,  /* T * */
        0, 2, 0,  /* struct; so that ARGB keeps stack aligned */
        1,        /* little_endian */
        1,        /* mulops_calls */
        0,        /* wants_callb */
        1,        /* wants_argb */
        0,        /* left_to_right */
        0,        /* wants_dag */
        address,
        blockbeg,
        blockend,
        defaddress,
        defconst,
        defstring,
        defsymbol,
        emit,
        export,
        function,
        gen,
        global,
        import,
        local,
        progbeg,
        progend,
        segment,
        space,
        0, 0, 0, 0, 0, 0, 0,
        {1, rmap,
            blkfetch, blkstore, blkloop,
            _label,
            _rule,
            _nts,
            _kids,
            _string,
            _templates,
            _isinstruction,
            _ntname,
            emit2,
            doarg,
            target,
            clobber,
}
};

